from .alifestd_is_asexual import alifestd_is_asexual
from .alifestd_is_sexual import alifestd_is_sexual
from .alifestd_make_ancestor_id_col import alifestd_make_ancestor_id_col
from .alifestd_make_ancestor_list_col import alifestd_make_ancestor_list_col
from .try_alifestd_validate import _try_alifestd_validate
from .all_unique import all_unique
from .is_subset import is_subset
from .parse_ancestor_list import parse_ancestor_list
from .rgetattr import rgetattr



import opytional as opyt
import pandas as pd
import typing

from ._impl import parse_ancestor_list as _parse_ancestor_list


def alife_dataframe_to_dict_of_lists(
    df: pd.DataFrame,
) -> typing.Dict[int, typing.List[int]]:
    """Extract an adjacency representation for an alife standard dataframe.

    Returns
    -------
    dict of int : list of int
        A dictionary where the keys are organism IDs and the values are the
        lists of organisms' ancestors.
    """

    df = df.copy()

    df['parsed_ancestor_list'] \
        = df['ancestor_list'].apply(_parse_ancestor_list)

    return df.set_index('id', drop=False)['parsed_ancestor_list'].to_dict()



#!/usr/bin/env python

'''
`anytree_tree_to_alife_dataframe` tests for
`alifedata-phyloinformatics-convert` package.
'''

import anytree
from iterpop import iterpop as ip
from os.path import dirname, realpath
import pandas as pd

import alifedata_phyloinformatics_convert as apc


def test_singleton():
    original_tree = anytree.Node("beep")
    assert original_tree.name == "beep"
    converted_df = apc.anytree_tree_to_alife_dataframe(original_tree)

    expected_df = pd.DataFrame({
        "id" : [0],
        "ancestor_list" : ["[None]"],
        "name" : ["beep"],
    })

    assert converted_df.equals(expected_df)


def test_simple_tree():
    udo = anytree.Node("Udo", id=0)
    marc = anytree.Node("Marc", parent=udo, id=1111)
    lian = anytree.Node("Lian", parent=marc, id=2222)
    dan = anytree.Node("Dan", parent=udo, id=333)
    jet = anytree.Node("Jet", parent=dan, id=444)
    jan = anytree.Node("Jan", parent=dan, id=555)
    joe = anytree.Node("Joe", parent=dan, id=666)

    converted_df = apc.anytree_tree_to_alife_dataframe(udo)
    expected_df = pd.DataFrame({
        "id" : [0, 1111, 333, 2222, 444, 555, 666],
        "ancestor_list" : [
            "[None]", "[0]", "[0]", "[1111]", "[333]", "[333]", "[333]"
        ],
        "name" : ["Udo", "Marc", "Dan", "Lian", "Jet", "Jan", "Joe"],
    })

    assert converted_df.equals(expected_df)


def test_simple_tree_origin_times():
    udo = anytree.Node("Udo", id=0, edge_length=1)
    marc = anytree.Node("Marc", parent=udo, id=1111, edge_length=1)
    lian = anytree.Node("Lian", parent=marc, id=2222, edge_length=1)
    dan = anytree.Node("Dan", parent=udo, id=333, edge_length=2)
    jet = anytree.Node("Jet", parent=dan, id=444, edge_length=4)
    jan = anytree.Node("Jan", parent=dan, id=555, edge_length=5)
    joe = anytree.Node("Joe", parent=dan, id=666, edge_length=6)

    converted_df = apc.anytree_tree_to_alife_dataframe(udo)
    expected_df = pd.DataFrame({
        "id" : [0, 1111, 333, 2222, 444, 555, 666],
        "ancestor_list" : [
            "[None]", "[0]", "[0]", "[1111]", "[333]", "[333]", "[333]"
        ],
        "edge_length" : [1, 1, 2, 1, 4, 5, 6],
        "name" : ["Udo", "Marc", "Dan", "Lian", "Jet", "Jan", "Joe"],
        "origin_time" : [1, 2, 3, 3, 7, 8, 9],
    })

    assert converted_df.equals(expected_df)



#!/usr/bin/env python

'''
`alife_dataframe_to_dict_of_lists` tests for
`alifedata-phyloinformatics-convert` package.
'''

import networkx as nx
import pandas as pd

import alifedata_phyloinformatics_convert as apc

def test_alife_dataframe_to_network_digraph_empty():
    df = pd.DataFrame({'id': [], 'ancestor_list': []})
    expected_output = nx.DiGraph()
    assert nx.is_isomorphic(
        apc.alife_dataframe_to_networkx_digraph(df),
        expected_output,
        node_match=True,
        edge_match=True,
    )

def test_alife_dataframe_to_network_digraph_asexual():
    df = pd.DataFrame(
        {
            'id': [0, 1, 7, 9],
            'ancestor_list': ['[None]', '[0]', '[1]', '[1]'],
            'taxon_label': ['a', 'b', 'c', 'd'],
        },
    )
    expected_output = nx.DiGraph(
        {0: [], 1: [0], 7: [1], 9: [1]}
    )

    for df_ in df, df.sample(frac=1):
        result = apc.alife_dataframe_to_networkx_digraph(df_)
        assert nx.to_dict_of_lists(expected_output) == nx.to_dict_of_lists(
            result
        )
        assert result.nodes[0]["taxon_label"] == "a"
        assert result.nodes[1]["taxon_label"] == "b"
        assert result.nodes[7]["taxon_label"] == "c"
        assert result.nodes[9]["taxon_label"] == "d"


def test_alife_dataframe_to_network_digraph_asexual_origin_time():
    df = pd.DataFrame(
        {
            'id': [0, 1, 7, 9],
            'ancestor_list': ['[None]', '[0]', '[1]', '[1]'],
            'origin_time': [0, 1, 12, 12],
        },
    )
    for df_ in df, df.sample(frac=1):
        result = apc.alife_dataframe_to_networkx_digraph(
            df_, setup_edge_lengths=True
        )
        assert result[1][0]["length"] == 1
        assert result[7][1]["length"] == 11
        assert result[9][1]["length"] == 11

        result = apc.alife_dataframe_to_networkx_digraph(
            df_, setup_edge_lengths=False
        )
        assert "length" not in result[1][0]
        assert "length" not in result[7][1]
        assert "length" not in result[9][1]


def test_alife_dataframe_to_network_digraph_asexual_edge_length():
    df = pd.DataFrame(
        {
            'id': [0, 1, 7, 9],
            'ancestor_list': ['[None]', '[0]', '[1]', '[1]'],
            'edge_length': [0, 1, 12, 12],
        },
    )
    for df_ in df, df.sample(frac=1):
        result = apc.alife_dataframe_to_networkx_digraph(
            df_, setup_edge_lengths=True
        )
        assert result[1][0]["length"] == 1
        assert result[7][1]["length"] == 12
        assert result[9][1]["length"] == 12

        result = apc.alife_dataframe_to_networkx_digraph(
            df_, setup_edge_lengths=False
        )
        assert "length" not in result[1][0]
        assert "length" not in result[7][1]
        assert "length" not in result[9][1]


def test_alife_dataframe_to_network_digraph_sexual():
    df = pd.DataFrame(
        {
            'id': [0, 1, 7, 9],
            'ancestor_list': ['[None]', '[None]', '[1,0]', '[1,7]'],
            'taxon_label': ['a', 'b', 'c', 'd'],
        }
    )

    expected_output = nx.DiGraph(
        {0: [], 1: [], 7: [1, 0], 9: [1, 7]}
    )

    for df_ in df, df.sample(frac=1):
        result = apc.alife_dataframe_to_networkx_digraph(df_)
        assert nx.to_dict_of_lists(expected_output) == nx.to_dict_of_lists(
            result
        )
        assert result.nodes[0]["taxon_label"] == "a"
        assert result.nodes[1]["taxon_label"] == "b"
        assert result.nodes[7]["taxon_label"] == "c"
        assert result.nodes[9]["taxon_label"] == "d"



# This file has been copied from hstrat
# <https://github.com/mmore500/hstrat/blob/47fe9048d6b327ca40f77424e3cf9392f7980689/hstrat/_auxiliary_lib/_all_unique.py>.
# It is up to date as of commit
# [1ce7a0e](https://github.com/mmore500/hstrat/commit/1ce7a0e).
# Eventually, it should be migrated to a standalone library.
import numpy as np


def all_unique(array: np.array) -> bool:
    """Are all values in `array` unique?"""
    seen_unique = set()
    for idx, val in enumerate(array):
        seen_unique.add(val)
        if idx + 1 != len(seen_unique):
            return False

    return True



# This file has been copied from hstrat
# <https://github.com/mmore500/hstrat/blob/47fe9048d6b327ca40f77424e3cf9392f7980689/hstrat/_auxiliary_lib/_alifestd_is_sexual.py>.
# It is up to date as of commit
# [1ce7a0e](https://github.com/mmore500/hstrat/commit/1ce7a0e).
# Eventually, it should be migrated to a standalone library.
import pandas as pd


def alifestd_is_sexual(phylogeny_df: pd.DataFrame) -> bool:
    """Do any organisms in the phylogeny have than one immediate ancestor?
    Input dataframe is not mutated by this operation.
    """
    return phylogeny_df["ancestor_list"].str.contains(",").any()



import functools


# adapted from https://stackoverflow.com/a/31174427/17332200
def rgetattr(obj, attr, *args):
    def _getattr(obj, attr):
        return getattr(obj, attr, *args)
    return functools.reduce(_getattr, [obj] + attr.split('.'))



import string
import typing


def parse_ancestor_list(raw: str) -> typing.List[int]:
    assert not any(
        c in raw for c in string.whitespace
    ), "Whitespace separated ancestor list not supported."
    try:
        ancestor_list = eval(raw)
        if ancestor_list == [None]:
            return []
        else:
            assert None not in ancestor_list
            return ancestor_list
    except NameError:
        # if ancestor list contains a placeholder string like NONE,
        # it will trigger a NameError when eval'ed
        return []



#!/usr/bin/env python

'''
`dendropy_tree_to_alife_dataframe` tests for
`alifedata-phyloinformatics-convert` package.
'''

import dendropy
from dendropy.calculate import treecompare
from iterpop import iterpop as ip
import networkx as nx
import numpy as np
from os.path import dirname, realpath

import alifedata_phyloinformatics_convert as apc


def test_newick1():

    source_tree = dendropy.Tree.get(
        path=f'{dirname(realpath(__file__))}/assets/APG_Angiosperms.newick',
        schema='newick',
    )

    source_df = apc.dendropy_tree_to_alife_dataframe(source_tree).dropna(
        axis=1, how='all'
    )
    networkx_tree = apc.alife_dataframe_to_networkx_digraph(source_df)

    reconverted_df = apc.networkx_digraph_to_alife_dataframe(networkx_tree)

    assert reconverted_df.drop(
        "ancestor_id", axis=1
    ).sort_values(by="id").sort_index(axis=1).reset_index(drop=True).equals(
        source_df.fillna(value=np.nan).sort_values(by="id").sort_index(axis=1).reset_index(
            drop=True
        )
    )


def test_newick2():

    source_tree = dendropy.Tree.get(
        path=f'{dirname(realpath(__file__))}/assets/APG_Angiosperms.newick',
        schema='newick',
    )
    source_tree.resolve_polytomies()
    source_tree.suppress_unifurcations()
    source_tree.collapse_basal_bifurcation()

    source_df = apc.dendropy_tree_to_alife_dataframe(source_tree).dropna(
        axis=1, how='all'
    )
    networkx_tree = apc.alife_dataframe_to_networkx_digraph(source_df)

    reconverted_df = apc.networkx_digraph_to_alife_dataframe(networkx_tree)

    assert reconverted_df.drop(
        "ancestor_id", axis=1
    ).sort_values(by="id").sort_index(axis=1).reset_index(drop=True).equals(
        source_df.fillna(value=np.nan).sort_values(
            by="id"
        ).sort_index(axis=1).reset_index(
            drop=True
        )
    )


def test_nexml():

    script_directory = dirname(realpath(__file__))
    source_tree = dendropy.Tree.get(
        path=f'{script_directory}/assets/pythonidae.annotated.nexml',
        schema='nexml',
    )
    source_df = apc.dendropy_tree_to_alife_dataframe(source_tree).dropna(
        axis=1, how='all'
    )

    networkx_tree = apc.alife_dataframe_to_networkx_digraph(source_df)
    reconverted_df = apc.networkx_digraph_to_alife_dataframe(networkx_tree)

    assert reconverted_df.drop(
        "ancestor_id", axis=1
    ).sort_values(by="id").sort_index(axis=1).reset_index(drop=True).equals(
        source_df.fillna(value=np.nan).sort_values(
            by="id"
        ).sort_index(axis=1).reset_index(
            drop=True
        )
    )


def test_relabel_ints():

    script_directory = dirname(realpath(__file__))
    source_tree = dendropy.Tree.get(
        path=f'{script_directory}/assets/pythonidae.annotated.nexml',
        schema='nexml',
    )
    source_df = apc.dendropy_tree_to_alife_dataframe(source_tree).dropna(
        axis=1, how='all'
    )

    networkx_tree = apc.alife_dataframe_to_networkx_digraph(source_df)
    nx.relabel_nodes(
        networkx_tree, {n : f"{n}!" for n in networkx_tree.nodes}, copy=False
    )
    reconverted_df = apc.networkx_digraph_to_alife_dataframe(networkx_tree)

    source_df["label"] = source_df["id"].apply(lambda x: f"{x}!")

    assert reconverted_df.drop(
        "ancestor_id", axis=1
    ).sort_values(by="id").sort_index(axis=1).reset_index(drop=True).equals(
        source_df.fillna(value=np.nan).sort_values(
            by="id"
        ).sort_index(axis=1).reset_index(
            drop=True
        )
    )



#!/usr/bin/env python

'''
`alife_dataframe_to_dict_of_lists` tests for
`alifedata-phyloinformatics-convert` package.
'''

import pandas as pd

import alifedata_phyloinformatics_convert as apc

def test_alife_dataframe_to_dict_of_lists_empty():
    df = pd.DataFrame({'id': [], 'ancestor_list': []})
    expected_output = {}
    assert apc.alife_dataframe_to_dict_of_lists(df) == expected_output

def test_alife_dataframe_to_dict_of_lists_asexual():
    df = pd.DataFrame(
        {'id': [0, 1, 7, 9], 'ancestor_list': ['[None]', '[0]', '[1]', '[1]']}
    )
    expected_output = {0: [], 1: [0], 7: [1], 9: [1]}
    assert apc.alife_dataframe_to_dict_of_lists(df) == expected_output
    assert apc.alife_dataframe_to_dict_of_lists(
        df.sample(frac=1)
    ) == expected_output

def test_alife_dataframe_to_dict_of_lists_sexual():
    df = pd.DataFrame(
        {
            'id': [0, 1, 7, 9],
            'ancestor_list': ['[None]', '[None]', '[1,0]', '[1,7]']
        }
    )

    expected_output = {0: [], 1: [], 7: [1, 0], 9: [1, 7]}
    assert apc.alife_dataframe_to_dict_of_lists(df) == expected_output
    assert apc.alife_dataframe_to_dict_of_lists(
        df.sample(frac=1)
    ) == expected_output



# This file has been copied from hstrat
# <https://github.com/mmore500/hstrat/blob/47fe9048d6b327ca40f77424e3cf9392f7980689/hstrat/_auxiliary_lib/_is_subset.py>.
# It is up to date as of commit
# [1ce7a0e](https://github.com/mmore500/hstrat/commit/1ce7a0e).
# Eventually, it should be migrated to a standalone library.
import numpy as np


def is_subset(subset: np.array, superset: np.array) -> bool:
    """Are all values in `subset` contained in `superset`?"""
    superset_lookup = set(superset)
    for val in subset:
        if val not in superset_lookup:
            return False

    return True



from Bio import Phylo
from collections import defaultdict
from collections.abc import Mapping
import functools
from iterpop import iterpop as ip
import itertools as it
from nanto import isanan
import networkx as nx
import numbers
import numpy as np
import opytional as opyt
import pandas as pd
import typing

from ._impl import rgetattr as _rgetattr


def networkx_digraph_to_alife_dataframe(
    graph: nx.Graph,
) -> pd.DataFrame:
    """Convert a networkx digraph to a dataframe formatted to the artificial
    life community data format standards.

    The following node attributes will automatically be exported to
    dataframe columns, if available:
        * branch_length,
        * edge_length,
        * length,
        * weight,
        * label,
        * name,
        * origin_time, and
        * taxon_label.

    Parameters
    ----------
    graph:
        networkx graph to convert.
    """

    if not all(
        isinstance(x, numbers.Integral) and x >= 0 for x in graph.nodes
    ):
        assert not any("label" in data for __, data in graph.nodes(data=True))
        graph = nx.convert_node_labels_to_integers(
            graph,
            label_attribute="label",
        )

    records = []
    visited_nodes = set(graph.nodes)
    for from_, to, edge_data in it.chain(
        graph.edges(data=True),
        (
            (ip.popsingleton(visited_nodes), None, {})
            for __ in range(1)
        ),
    ):
        attrs = graph.nodes[from_]
        def try_get_attr(attr):
            return opyt.or_else(
                attrs.get(attr, None),
                lambda: edge_data.get(attr, np.nan),
            )

        record = {
            "id": from_,
            "ancestor_list" : str([to]),
            "ancestor_id" : opyt.or_value(to, from_),
            "branch_length" : try_get_attr("branch_length"),
            "edge_length" : try_get_attr("edge_length"),
            "length" : try_get_attr("length"),
            "weight" : try_get_attr("weight"),
            "label" : try_get_attr("label"),
            "name" : try_get_attr("name"),
            "origin_time" : try_get_attr("origin_time"),
            "taxon_label" : try_get_attr("taxon_label"),
        }
        visited_nodes.remove(from_)
        records.append(record)

    return pd.DataFrame.from_records(records).dropna(axis=1, how='all')



# This file has been copied from hstrat
# <https://github.com/mmore500/hstrat/blob/47fe9048d6b327ca40f77424e3cf9392f7980689/hstrat/_auxiliary_lib/_alifestd_make_ancestor_list_col.py>.
# It is up to date as of commit
# [1ce7a0e](https://github.com/mmore500/hstrat/commit/1ce7a0e).
# Eventually, it should be migrated to a standalone library.

import pandas as pd


def alifestd_make_ancestor_list_col(
    ids: pd.Series, ancestor_ids: pd.Series
) -> pd.Series:
    """Translate a column of integer ancestor id values into alife standard
    `ancestor_list` representation."""

    res = ancestor_ids.map("[{}]".format)
    res[ids == ancestor_ids] = "[none]"

    return



#!/usr/bin/env python

"""
`RosettaTree` tests for `alifedata-phyloinformatics-convert` package.
"""

from os.path import dirname, realpath

import anytree
from Bio import Phylo as BioPhylo
import dendropy as dp
import networkx as nx
import pandas as pd
import pytest

import alifedata_phyloinformatics_convert as apc


@pytest.mark.parametrize(
    "original_df",
    [
        pd.read_csv(f"{dirname(realpath(__file__))}/assets/alifedata.csv"),
        pd.read_csv(
            f"{dirname(realpath(__file__))}/assets/alifedata_minimal.csv",
        ),
    ],
)
def test_alife_to_biopython(original_df):
    expected_tree = apc.alife_dataframe_to_biopython_tree(
        original_df, setup_branch_lengths=True
    )
    rosetta_tree = apc.RosettaTree(original_df)

    # twice to test caching
    for __ in range(2):
        converted_tree = rosetta_tree.as_biopython
        assert str(converted_tree) == str(expected_tree)


@pytest.mark.parametrize(
    "original_df",
    [
        pd.read_csv(f"{dirname(realpath(__file__))}/assets/alifedata.csv"),
        pd.read_csv(
            f"{dirname(realpath(__file__))}/assets/alifedata_minimal.csv",
        ),
    ],
)
def test_alife_to_dendropy(original_df):
    expected_tree = apc.alife_dataframe_to_dendropy_tree(
        original_df, setup_edge_lengths=True
    )
    rosetta_tree = apc.RosettaTree(original_df)

    # twice to test caching
    for __ in range(2):
        converted_tree = rosetta_tree.as_dendropy
        assert str(converted_tree) == str(expected_tree)


@pytest.mark.parametrize(
    "original_df",
    [
        pd.read_csv(f"{dirname(realpath(__file__))}/assets/alifedata.csv"),
        pd.read_csv(
            f"{dirname(realpath(__file__))}/assets/alifedata_minimal.csv",
        ),
    ],
)
def test_alife_to_networkx(original_df):
    expected_tree = apc.alife_dataframe_to_networkx_digraph(
        original_df, setup_edge_lengths=True
    )
    rosetta_tree = apc.RosettaTree(original_df)

    # twice to test caching
    for __ in range(2):
        converted_tree = rosetta_tree.as_networkx
        assert "\n".join(
            nx.generate_gml(converted_tree)
        ) == "\n".join(
            nx.generate_gml(expected_tree)
        )


@pytest.mark.parametrize(
    "original_df",
    [
        pd.read_csv(f"{dirname(realpath(__file__))}/assets/alifedata.csv"),
        pd.read_csv(
            f"{dirname(realpath(__file__))}/assets/alifedata_minimal.csv",
        ),
    ],
)
def test_biopython_to_alife(original_df):
    original_tree = apc.alife_dataframe_to_biopython_tree(
        original_df, setup_branch_lengths=True
    )
    expected_df = apc.biopython_tree_to_alife_dataframe(
        original_tree, {"name": "taxon_label"}
    )
    rosetta_tree = apc.RosettaTree(original_tree)

    # twice to test caching
    for __ in range(2):
        converted_df = rosetta_tree.as_alife
        assert converted_df.equals(expected_df)


@pytest.mark.parametrize(
    "original_df",
    [
        pd.read_csv(f"{dirname(realpath(__file__))}/assets/alifedata.csv"),
        pd.read_csv(
            f"{dirname(realpath(__file__))}/assets/alifedata_minimal.csv",
        ),
    ],
)
def test_biopython_to_dendropy(original_df):
    original_tree = apc.alife_dataframe_to_biopython_tree(
        original_df, setup_branch_lengths=True
    )
    expected_tree = apc.alife_dataframe_to_dendropy_tree(
        original_df, setup_edge_lengths=True
    )
    rosetta_tree = apc.RosettaTree(original_tree)

    # twice to test caching
    for __ in range(2):
        converted_tree = rosetta_tree.as_dendropy
        assert str(converted_tree) == str(expected_tree)


@pytest.mark.parametrize(
    "original_df",
    [
        pd.read_csv(f"{dirname(realpath(__file__))}/assets/alifedata.csv"),
        pd.read_csv(
            f"{dirname(realpath(__file__))}/assets/alifedata_minimal.csv",
        ),
    ],
)
def test_dendropy_to_alife(original_df):
    original_tree = apc.alife_dataframe_to_dendropy_tree(
        original_df, setup_edge_lengths=True
    )
    expected_df = apc.dendropy_tree_to_alife_dataframe(original_tree)
    rosetta_tree = apc.RosettaTree(original_tree)

    # twice to test caching
    for __ in range(2):
        converted_df = rosetta_tree.as_alife
        assert converted_df.equals(expected_df)


@pytest.mark.parametrize(
    "original_df",
    [
        pd.read_csv(f"{dirname(realpath(__file__))}/assets/alifedata.csv"),
        pd.read_csv(
            f"{dirname(realpath(__file__))}/assets/alifedata_minimal.csv",
        ),
    ],
)
def test_dendropy_to_biopython(original_df):
    original_tree = apc.alife_dataframe_to_dendropy_tree(
        original_df, setup_edge_lengths=True
    )
    expected_tree = apc.alife_dataframe_to_biopython_tree(
        original_df, setup_branch_lengths=True
    )
    rosetta_tree = apc.RosettaTree(original_tree)

    # twice to test caching
    for __ in range(2):
        converted_tree = rosetta_tree.as_biopython
        assert str(converted_tree) == str(expected_tree)



@pytest.mark.parametrize(
    "original_df",
    [
        pd.read_csv(f"{dirname(realpath(__file__))}/assets/alifedata.csv"),
        pd.read_csv(
            f"{dirname(realpath(__file__))}/assets/alifedata_minimal.csv",
        ),
    ],
)
def test_networkx_to_alife(original_df):
    original_tree = apc.alife_dataframe_to_networkx_digraph(
        original_df, setup_edge_lengths=True
    )
    expected_df = apc.networkx_digraph_to_alife_dataframe(original_tree)
    rosetta_tree = apc.RosettaTree(original_tree)

    # twice to test caching
    for __ in range(2):
        converted_df = rosetta_tree.as_alife
        assert converted_df.equals(expected_df)


@pytest.mark.parametrize(
    "original_df",
    [
        pd.read_csv(f"{dirname(realpath(__file__))}/assets/alifedata.csv"),
        pd.read_csv(
            f"{dirname(realpath(__file__))}/assets/alifedata_minimal.csv",
        ),
    ],
)
def test_networkx_to_biopython(original_df):
    original_tree = apc.alife_dataframe_to_networkx_digraph(
        original_df, setup_edge_lengths=True
    )
    expected_tree = apc.alife_dataframe_to_biopython_tree(
        original_df, setup_branch_lengths=True
    )
    rosetta_tree = apc.RosettaTree(original_tree)

    # twice to test caching
    for __ in range(2):
        converted_tree = rosetta_tree.as_biopython
        assert str(converted_tree) == str(expected_tree)


@pytest.mark.parametrize(
    "original_df",
    [
        pd.read_csv(f"{dirname(realpath(__file__))}/assets/alifedata.csv"),
        pd.read_csv(
            f"{dirname(realpath(__file__))}/assets/alifedata_minimal.csv",
        ),
    ],
)
def test_networkx_to_dendropy(original_df):
    original_tree = apc.alife_dataframe_to_networkx_digraph(
        original_df, setup_edge_lengths=True
    )
    expected_tree = apc.alife_dataframe_to_dendropy_tree(
        original_df, setup_edge_lengths=True
    )
    rosetta_tree = apc.RosettaTree(original_tree)

    # twice to test caching
    for __ in range(2):
        converted_tree = rosetta_tree.as_dendropy
        assert str(converted_tree) == str(expected_tree)



"""Top-level package for alifedata-phyloinformatics-convert."""

__author__ = """Matthew Andres Moreno"""
__email__ = 'm.more500@gmail.com'
__version__ = '0.12.0'

from .alife_dataframe_to_biopython_tree \
    import alife_dataframe_to_biopython_tree
from .alife_dataframe_to_biopython_trees \
    import alife_dataframe_to_biopython_trees
from .alife_dataframe_to_dict_of_lists import alife_dataframe_to_dict_of_lists
from .alife_dataframe_to_dendropy_tree import alife_dataframe_to_dendropy_tree
from .alife_dataframe_to_dendropy_trees \
    import alife_dataframe_to_dendropy_trees
from .alife_dataframe_to_networkx_digraph \
    import alife_dataframe_to_networkx_digraph
from .anytree_tree_to_alife_dataframe \
    import anytree_tree_to_alife_dataframe
from .biopython_tree_to_alife_dataframe \
    import biopython_tree_to_alife_dataframe
from .dendropy_tree_to_alife_dataframe import dendropy_tree_to_alife_dataframe
from .networkx_digraph_to_alife_dataframe \
    import networkx_digraph_to_alife_dataframe
from .dendropy_tree_to_scipy_linkage_matrix \
    import dendropy_tree_to_scipy_linkage_matrix
from .scipy_linkage_matrix_to_dendropy_tree \
    import scipy_linkage_matrix_to_dendropy_tree
from .RosettaTree import RosettaTree

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    'alife_dataframe_to_biopython_tree',
    'alife_dataframe_to_biopython_trees',
    'alife_dataframe_to_dendropy_tree',
    'alife_dataframe_to_dendropy_trees',
    'alife_dataframe_to_networkx_digraph',
    'anytree_tree_to_alife_dataframe',
    'biopython_tree_to_alife_dataframe',
    'dendropy_tree_to_alife_dataframe',
    'dendropy_tree_to_scipy_linkage_matrix',
    'networkx_digraph_to_alife_dataframe',
    'scipy_linkage_matrix_to_dendropy_tree',
    'RosettaTree',
]



import anytree
from collections.abc import Mapping
import functools
from nanto import isanan
import opytional as opyt
import pandas as pd
import typing

from ._impl import rgetattr as _rgetattr


def anytree_tree_to_alife_dataframe(
    tree: anytree.AnyNode,
) -> pd.DataFrame:
    """Convert a anytree tree to a dataframe formatted to the
    artificial life community data format standards.

    The following Node object attributes will automatically be exported to
    dataframe columns, if available:
        * edge_length,
        * id,
        * label,
        * name,
        * origin_time, and
        * taxon_label.

    Parameters
    ----------
    tree:
        anytree tree to convert.
    """

    # set up node origin times if any edge lengths set
    if any(
        getattr(node, 'edge_length', None) is not None
        for node in anytree.PreOrderIter(tree)
    ):
        if not hasattr(tree, 'origin_time'):
            tree.origin_time = opyt.or_value(
                getattr(tree, "edge_length", None),
                0
            )
        else:
            assert tree.origin_time is not None
        for node in anytree.LevelOrderIter(tree):
            parent = node.parent
            if parent is not None and not hasattr(node, 'origin_time'):
                if None not in (
                    getattr(parent, 'origin_time', None),
                    getattr(node, 'edge_length', None),
                ) and (
                    not isanan(parent.origin_time)
                    and not isanan(node.edge_length)
                ):
                    node.origin_time = parent.origin_time + node.edge_length

    # attach ids to nodes, if needed
    for fallback_id, node in enumerate(anytree.LevelOrderIter(tree)):
        if not hasattr(node, 'id'):
            node.id = fallback_id
        else:
            assert isinstance(node.id, int)

    return pd.DataFrame.from_records([
        {
            'id': node.id,
            'ancestor_list': f"[{opyt.apply_if(node.parent, lambda x: x.id)}]",
            'edge_length': getattr(node, 'edge_length', None),
            'label': getattr(node, 'label', None),
            'name': getattr(node, 'name', None),
            'origin_time': getattr(node, 'origin_time', None),
            'taxon_label': getattr(node, 'taxon_label', None),
        }
        for node in anytree.LevelOrderIter(tree)
    ]).dropna(axis=1, how="all")



import networkx as nx
import pandas as pd
import typing

from .alife_dataframe_to_dict_of_lists import alife_dataframe_to_dict_of_lists


# see also https://github.com/alife-data-standards/alife-std-dev-python/blob/f21a63d70077345441b9b52c3f470f5dca1127c1/tests/test_phylogeny_loader.py
def alife_dataframe_to_networkx_digraph(
    df: pd.DataFrame,
    setup_edge_lengths: bool = False,
) -> nx.DiGraph:
    """Open a phylogeny dataframe formatted to the artificial life community
    data format standards as a networkx directed graph.

    Directed edges point from child to parent. Clades that do not share a
    common ancestor are supported. Call `.reverse()` to orient so directed
    edges point from parent to child.

    If enabled, branch lengths will be set up based on the origin_time
    attribute.

    The following column values will automatically be applied as node attributes, if available:
        * branch_length,
        * edge_length,
        * length,
        * weight,
        * label,
        * name,
        * origin_time, and
        * taxon_label.

    Will raise ValueError if nonequivalent branch_length and edge_length
    columns are provided.

    Parameters
    ----------
    df:
        Pandas DataFrame to convert.
    setup_edge_lengths: bool, optional
        Should we try to set up edge lengths using the origin_time column?
        Will not override if branch_length or edge_length is provided as a
        column of df.
    """

    if "branch_length" in df and "edge_length" in df and not (
        df["branch_length"].equals(df["edge_length"])
    ):
        raise ValueError

    g = nx.from_dict_of_lists(
        alife_dataframe_to_dict_of_lists(df),
        create_using=nx.DiGraph,
    )

    nx.set_node_attributes(
        g,
        df.set_index("id")[[
            attr
            for attr in (
                "branch_length",
                "edge_length",
                "length",
                "weight",
                "label",
                "name",
                "origin_time",
                "taxon_label",
            )
            if attr in df
        ]].to_dict(orient="index"),
    )

    if setup_edge_lengths and "length" in df:
        for from_, to in g.edges:
            g[from_][to]["length"] = g.nodes[from_]["length"]
    elif setup_edge_lengths and "edge_length" in df:
        for from_, to in g.edges:
            g[from_][to]["length"] = g.nodes[from_]["edge_length"]
    elif setup_edge_lengths and "branch_length" in df:
        for from_, to in g.edges:
            g[from_][to]["length"] = g.nodes[from_]["branch_length"]
    elif setup_edge_lengths and "origin_time" in df:
        for from_, to in g.edges:
            g[from_][to]["length"] = (
                g.nodes[from_]["origin_time"] - g.nodes[to]["origin_time"]
            )
    elif setup_edge_lengths:
        for from_, to in g.edges:
            g[from_][to]["length"] = 1

    return g



import Bio
import dendropy
from functools import lru_cache
import networkx as nx
import pandas
import typing

from ._impl import _try_alifestd_validate as _try_alifestd_validate

from .alife_dataframe_to_biopython_tree \
    import alife_dataframe_to_biopython_tree
from .alife_dataframe_to_dendropy_tree \
    import alife_dataframe_to_dendropy_tree
from .alife_dataframe_to_networkx_digraph \
    import alife_dataframe_to_networkx_digraph
from .biopython_tree_to_alife_dataframe \
    import biopython_tree_to_alife_dataframe
from .dendropy_tree_to_alife_dataframe \
    import dendropy_tree_to_alife_dataframe
from .networkx_digraph_to_alife_dataframe \
    import networkx_digraph_to_alife_dataframe


class RosettaTree:
    """Adapter class for implicit conversion between tree representations
    across phylogenetic libraries.

    Enables user to translate across tree implementations without repeatedly
    calling the individual conversion functions. This allows users to accept
    any such tree as an argument in a library-agnostic way.
    """

    _tree: pandas.DataFrame

    def __init__(self, tree: typing.Union[
        dendropy.Tree,
        nx.DiGraph,
        pandas.DataFrame,
        Bio.Phylo.BaseTree.Tree
    ]) -> None:
        """Set up underlying data to be viewed, from any supported format."""
        # convert any supported tree format to ALife format,
        # as this is our interal representation
        if isinstance(tree, dendropy.Tree):
            # is a Dendropy Tree
            self._tree = dendropy_tree_to_alife_dataframe(tree)
        elif isinstance(tree, Bio.Phylo.BaseTree.Tree):
            # is a biopython tree
            self._tree = biopython_tree_to_alife_dataframe(
                tree, {'name': 'taxon_label'}
            )
        elif isinstance(tree, nx.DiGraph):
            # is a networkx digraph
            self._tree = networkx_digraph_to_alife_dataframe(tree)
        elif (
            isinstance(tree, pandas.DataFrame)
            and len(tree)
            and _try_alifestd_validate(tree)
        ):
            # is an Alife Dataframe
            self._tree = tree
        else:
            raise ValueError("Unsupported tree format")

    @property
    @lru_cache(maxsize=None)
    def as_biopython(self: "RosettaTree") -> Bio.Phylo.BaseTree.Tree:
        """Return stored tree as a BioPython tree."""
        return alife_dataframe_to_biopython_tree(
            self._tree, setup_branch_lengths=True
        )

    @property
    @lru_cache(maxsize=None)
    def as_dendropy(self: "RosettaTree") -> dendropy.Tree:
        """Return stored tree as a DendroPy tree."""
        return alife_dataframe_to_dendropy_tree(
            self._tree, setup_edge_lengths=True
        )

    @property
    @lru_cache(maxsize=None)
    def as_networkx(self: "RosettaTree") -> nx.DiGraph:
        """Return stored tree as a NetWorkX DiGraph tree."""
        return alife_dataframe_to_networkx_digraph(
            self._tree, setup_edge_lengths=True
        )

    @property
    @lru_cache(maxsize=None)
    def as_alife(self: "RosettaTree") -> pandas.DataFrame:
        """Return stored tree as a dataframe in alife standard format."""
        return self._tree
